#include "binio.h"
#include "binio-internal.h"
#include <stdio.h>
#include <stdlib.h>

static
int
integers(void)
{
	signed char const a1s0 = -13, a1s1 = 49;
	unsigned char const a1u0 = 5, a1u1 = 205;
	short const a2s0 = -32740, a2s1 = 13099;
	unsigned short const a2u0 = 0, a2u1 = 65490u;
	int const a3s0 = -2, a3s1 = rand();
	unsigned int const a3u0 = rand(), a3u1 = 44509u;
	long const a4s0 = -rand(), a4s1 = 292818374l;
	unsigned long const a4u0 = rand(), a4u1 = 3918398471ul;
	long_long const a5s0 = -(rand() << 9), a5s1 = rand() + 217128743610;
	unsigned_long_long const a5u0 = rand() << 6, a5u1 = 9223372036856699789;
	
	signed char b1s0, b1s1;
	unsigned char b1u0, b1u1;
	short b2s0, b2s1;
	unsigned short b2u0, b2u1;
	int b3s0, b3s1;
	unsigned int b3u0, b3u1;
	long b4s0, b4s1;
	unsigned long b4u0, b4u1;
	long_long b5s0, b5s1;
	unsigned_long_long b5u0, b5u1;
	
	FILE * restrict f;
	



	if (! (f = fopen("binio-check.tmp", "wb"))) {
		fprintf(stderr, "could not open file for writing\n");
		return 1;
	}

#define A(x)	a ## x ## s0, a ## x ## s1, a ## x ## u0, a ## x ## u1
	fputf(f, "%hhd%hhd%hhu%hhu%hd%hd%hu%hu%d%d%u%u%ld%ld%lu%lu%lld%Ld%llu%Lu", A(1), A(2),
		A(3), A(4), A(5));
#undef A

	if (fclose(f))
		fprintf(stderr, "WARNING: error closing file properly\n");

	if (! (f = fopen("binio-check.tmp", "rb"))) {
		fprintf(stderr, "could not open file for reading\n");
		return 1;
	}

#define B(x)	&b ## x ## s0, &b ## x ## s1, &b ## x ## u0, &b ## x ## u1
	fgetf(f, "%hhd%hhd%hhu%hhu%hd%hd%hu%hu%d%d%u%u%ld%ld%lu%lu%lld%Ld%llu%Lu", B(1), B(2),
		B(3), B(4), B(5));
#undef B

	if (fclose(f))
		fprintf(stderr, "WARNING: error closing file properly\n");
	
	remove("binio-check.tmp");

#define EQ(x, t, n)	\
	((a ## x ## t ## n == b ## x ## t ## n) \
	|| (printf("%" INTMAX_T_FMT "d != %" INTMAX_T_FMT "d\n", (intmax_t)a ## x ## t ## n, \
		(intmax_t)b ## x ## t ## n), 0))
#define E(x)	(EQ(x, s, 0) && EQ(x, s, 1) && EQ(x, u, 0) && EQ(x, u, 1))
	if (E(1) && E(2) && E(3) && E(4) && E(5)) {
		printf("integer values matched correctly!\n");
		return 0;
	} else {
		fprintf(stderr, "mismatch between integer values\n");
		return 1;
	}
#undef E
#undef EQ
}

#include <math.h>

static
long_double
epsilon(long_double a, long_double b)
{
	if (b == 0.0l)
		return a;
	else if (a == 0.0l)
		return b;
	else {
		long_double n = a / b;
		return fabsl(1.0 - (n < 1.0 ? 1 / n : n));
	}
}

#include <float.h>

static
int
floats(void)
{
	float const a10 = -3.49e-1, a11 = 6.9381e12;
	double const a20 = 0.0, a21 = 3.9488399147266775e24;
	long_double const a30 = 4.9e-29, a31 = -9.928173672e6;
	
	float b10, b11;
	double b20, b21;
	long_double b30, b31;
	
	FILE * restrict f;
	



	if (! (f = fopen("binio-check.tmp", "wb"))) {
		fprintf(stderr, "could not open file for writing\n");
		return 1;
	}

#define A(x)	a ## x ## 0, a ## x ## 1
	fputf(f, "%f%f%f%f%Lf%Lf", A(1), A(2), A(3));
#undef A

	if (fclose(f))
		fprintf(stderr, "WARNING: error closing file properly\n");
	
	if (! (f = fopen("binio-check.tmp", "rb"))) {
		fprintf(stderr, "could not open file for reading\n");
		return 1;
	}

#define B(x)	&b ## x ## 0, &b ## x ## 1
	fgetf(f, "%f%f%lf%lf%Lf%Lf", B(1), B(2), B(3));
#undef B

	if (fclose(f))
		fprintf(stderr, "WARNING: error closing file properly\n");
	
	remove("binio-check.tmp");



#define EQ(x, n)	\
	((epsilon(a ## x ## n, b ## x ## n) < 1.0e-8l / (1ull << ((x) * 8))) || \
	(printf("%.20g != %.20g\n", (double)(a ## x ## n), (double)(b ## x ## n)), 0))
#define E(x)	(EQ(x, 0) && EQ(x, 1))
	if (E(1) && E(2) && E(3)) {
		printf("floating point values matched adequately!\n");
		return 0; 
	} else {
		fprintf(stderr, "unacceptable mismatch between floating point values\n");
		return 1;
	}
#undef E
#undef EQ
}

#include <time.h>

int
main(void)
{
	srand(time(0));

	if (integers() || floats()) {
		fprintf(stderr, "test failed!\n");
		return EXIT_FAILURE;
	}
	
	printf("test completed successfully!\n");
	
	return 0;
}
